import yasfpy.log as log

import numpy as np


class InitialField:
    def __init__(
        self,
        beam_width,
        focal_point,
        field_type: str = "gaussian",
        amplitude: float = 1,
        polar_angle: float = 0,
        azimuthal_angle: float = 0,
        polarization: str = "TE",
    ):
        self.field_type = field_type
        self.amplitude = amplitude
        self.polar_angle = polar_angle
        self.azimuthal_angle = azimuthal_angle
        self.polarization = polarization
        self.beam_width = beam_width
        self.focal_point = focal_point

        self.log = log.scattering_logger(__name__)
        self.__setup()

    def __set_pol_idx(self):
        if (
            isinstance(self.polarization, str) and self.polarization.lower() == "unp"
        ) or (isinstance(self.polarization, int) and self.polarization == 0):
            # Unpolarized is also present in the MSTM output
            self.pol = 0
        elif (
            isinstance(self.polarization, str) and self.polarization.lower() == "te"
        ) or (isinstance(self.polarization, int) and self.polarization == 1):
            # Coresponds to the perpendicular value found in MSTM
            self.pol = 1
        elif (
            isinstance(self.polarization, str) and self.polarization.lower() == "tm"
        ) or (isinstance(self.polarization, int) and self.polarization == 2):
            # Coresponds to the parallel value found in MSTM
            self.pol = 2
        else:
            self.pol = 0
            self.log.warning(
                "{} is not a valid polarization type. Please use TE or TM. Reverting to unpolarized".format(
                    self.polarization
                )
            )

    def __set_normal_incidence(self):
        self.normal_incidence = np.abs(np.sin(self.polar_angle)) < 1e-5

    def __setup(self):
        self.__set_pol_idx()
        self.__set_normal_incidence()



import logging
from typing import Union
from math import ceil

import numpy as np
from numba import cuda
from yasfpy.functions.spherical_functions_trigon import spherical_functions_trigon

from yasfpy.simulation import Simulation

from yasfpy.functions.cpu_numba import (
    compute_scattering_cross_section,
    compute_electric_field_angle_components,
    compute_polarization_components,
)
from yasfpy.functions.cuda_numba import (
    compute_scattering_cross_section_gpu,
    compute_electric_field_angle_components_gpu,
    compute_polarization_components_gpu,
)

# from yasfpy.functions.cpu_numba import compute_scattering_cross_section, compute_radial_independent_scattered_field, compute_electric_field_angle_components, compute_polarization_components
# from yasfpy.functions.cuda_numba import compute_scattering_cross_section_gpu, compute_radial_independent_scattered_field_gpu, compute_electric_field_angle_components_gpu, compute_polarization_components_gpu


class Optics:
    def __init__(self, simulation: Simulation):
        self.simulation = simulation

        self.c_ext = np.zeros_like(simulation.parameters.wavelength, dtype=complex)
        self.c_sca = np.zeros_like(simulation.parameters.wavelength, dtype=complex)

        self.log = logging.getLogger(__name__)

    def compute_cross_sections(self):
        a = self.simulation.initial_field_coefficients
        f = self.simulation.scattered_field_coefficients

        self.c_ext = np.zeros(
            self.simulation.parameters.wavelengths_number, dtype=complex
        )
        self.c_ext -= (
            np.sum(np.conjugate(a) * f, axis=(0, 1))
            / np.power(self.simulation.parameters.k_medium, 2)
            * np.pi
        )

        lmax = self.simulation.numerics.lmax
        particle_number = self.simulation.parameters.particles.number
        wavelengths = self.simulation.parameters.k_medium.shape[0]
        translation_table = self.simulation.numerics.translation_ab5
        associated_legendre_lookup = self.simulation.plm
        spherical_bessel_lookup = self.simulation.sph_j
        e_j_dm_phi_loopup = self.simulation.e_j_dm_phi

        idx_lookup = self.simulation.idx_lookup

        if self.simulation.numerics.gpu:
            c_sca_real = np.zeros(wavelengths, dtype=float)
            c_sca_imag = np.zeros_like(c_sca_real)

            idx_device = cuda.to_device(idx_lookup)
            sfc_device = cuda.to_device(f)
            c_sca_real_device = cuda.to_device(c_sca_real)
            c_sca_imag_device = cuda.to_device(c_sca_imag)
            translation_device = cuda.to_device(translation_table)
            associated_legendre_device = cuda.to_device(associated_legendre_lookup)
            spherical_bessel_device = cuda.to_device(spherical_bessel_lookup)
            e_j_dm_phi_device = cuda.to_device(e_j_dm_phi_loopup)

            jmax = particle_number * 2 * lmax * (lmax + 2)
            threads_per_block = (16, 16, 2)
            blocks_per_grid_x = ceil(jmax / threads_per_block[0])
            blocks_per_grid_y = ceil(jmax / threads_per_block[1])
            blocks_per_grid_z = ceil(wavelengths / threads_per_block[2])
            blocks_per_grid = (blocks_per_grid_x, blocks_per_grid_y, blocks_per_grid_z)

            compute_scattering_cross_section_gpu[blocks_per_grid, threads_per_block](
                lmax,
                particle_number,
                idx_device,
                sfc_device,
                translation_device,
                associated_legendre_device,
                spherical_bessel_device,
                e_j_dm_phi_device,
                c_sca_real_device,
                c_sca_imag_device,
            )
            c_sca_real = c_sca_real_device.copy_to_host()
            c_sca_imag = c_sca_imag_device.copy_to_host()
            c_sca = c_sca_real + 1j * c_sca_imag

        else:
            # from numba_progress import ProgressBar
            # num_iterations = jmax * jmax * wavelengths
            # progress = ProgressBar(total=num_iterations)
            progress = None
            c_sca = compute_scattering_cross_section(
                lmax,
                particle_number,
                idx_lookup,
                f,
                translation_table,
                associated_legendre_lookup,
                spherical_bessel_lookup,
                e_j_dm_phi_loopup,
                progress,
            )

        self.c_sca = (
            c_sca / np.power(np.abs(self.simulation.parameters.k_medium), 2) * np.pi
        )

        self.c_ext = np.real(self.c_ext)
        self.c_sca = np.real(self.c_sca)

        self.albedo = self.c_sca / self.c_ext

    def compute_phase_funcition(
        self,
        legendre_coefficients_number: int = 15,
        c_and_b: Union[bool, tuple] = False,
    ):
        pilm, taulm = spherical_functions_trigon(
            self.simulation.numerics.lmax, self.simulation.numerics.polar_angles
        )

        if self.simulation.numerics.gpu:
            jmax = (
                self.simulation.parameters.particles.number
                * self.simulation.numerics.nmax
            )
            angles = self.simulation.numerics.azimuthal_angles.size
            wavelengths = self.simulation.parameters.k_medium.size
            e_field_theta_real = np.zeros(
                (
                    self.simulation.numerics.azimuthal_angles.size,
                    self.simulation.parameters.k_medium.size,
                ),
                dtype=float,
            )
            e_field_theta_imag = np.zeros_like(e_field_theta_real)
            e_field_phi_real = np.zeros_like(e_field_theta_real)
            e_field_phi_imag = np.zeros_like(e_field_theta_real)

            particles_position_device = cuda.to_device(
                self.simulation.parameters.particles.position
            )
            idx_device = cuda.to_device(self.simulation.idx_lookup)
            sfc_device = cuda.to_device(self.simulation.scattered_field_coefficients)
            k_medium_device = cuda.to_device(self.simulation.parameters.k_medium)
            azimuthal_angles_device = cuda.to_device(
                self.simulation.numerics.azimuthal_angles
            )
            e_r_device = cuda.to_device(self.simulation.numerics.e_r)
            pilm_device = cuda.to_device(pilm)
            taulm_device = cuda.to_device(taulm)
            e_field_theta_real_device = cuda.to_device(e_field_theta_real)
            e_field_theta_imag_device = cuda.to_device(e_field_theta_imag)
            e_field_phi_real_device = cuda.to_device(e_field_phi_real)
            e_field_phi_imag_device = cuda.to_device(e_field_phi_imag)

            sizes = (jmax, angles, wavelengths)
            threads_per_block = (16, 16, 2)
            blocks_per_grid = tuple(
                [
                    ceil(sizes[k] / threads_per_block[k])
                    for k in range(len(threads_per_block))
                ]
            )
            # blocks_per_grid = (
            #   ceil(jmax / threads_per_block[0]),
            #   ceil(angles / threads_per_block[1]),
            #   ceil(wavelengths / threads_per_block[2]))

            compute_electric_field_angle_components_gpu[
                blocks_per_grid, threads_per_block
            ](
                self.simulation.numerics.lmax,
                particles_position_device,
                idx_device,
                sfc_device,
                k_medium_device,
                azimuthal_angles_device,
                e_r_device,
                pilm_device,
                taulm_device,
                e_field_theta_real_device,
                e_field_theta_imag_device,
                e_field_phi_real_device,
                e_field_phi_imag_device,
            )

            e_field_theta_real = e_field_theta_real_device.copy_to_host()
            e_field_theta_imag = e_field_theta_imag_device.copy_to_host()
            e_field_phi_real = e_field_phi_real_device.copy_to_host()
            e_field_phi_imag = e_field_phi_imag_device.copy_to_host()
            e_field_theta = e_field_theta_real + 1j * e_field_theta_imag
            e_field_phi = e_field_phi_real + 1j * e_field_phi_imag

            intensity = np.zeros_like(e_field_theta_real)
            dop = np.zeros_like(e_field_theta_real)
            dolp = np.zeros_like(e_field_theta_real)
            dolq = np.zeros_like(e_field_theta_real)
            dolu = np.zeros_like(e_field_theta_real)
            docp = np.zeros_like(e_field_theta_real)

            intensity_device = cuda.to_device(intensity)
            dop_device = cuda.to_device(dop)
            dolp_device = cuda.to_device(dolp)
            dolq_device = cuda.to_device(dolq)
            dolu_device = cuda.to_device(dolu)
            docp_device = cuda.to_device(docp)

            sizes = (angles, wavelengths)
            threads_per_block = (32, 32)
            blocks_per_grid = tuple(
                [
                    ceil(sizes[k] / threads_per_block[k])
                    for k in range(len(threads_per_block))
                ]
            )
            compute_polarization_components_gpu[blocks_per_grid, threads_per_block](
                self.simulation.parameters.k_medium.size,
                self.simulation.numerics.azimuthal_angles.size,
                e_field_theta_real_device,
                e_field_theta_imag_device,
                e_field_phi_real_device,
                e_field_phi_imag_device,
                intensity_device,
                dop_device,
                dolp_device,
                dolq_device,
                dolu_device,
                docp_device,
            )

            intensity = intensity_device.copy_to_host()
            dop = dop_device.copy_to_host()
            dolp = dolp_device.copy_to_host()
            dolq = dolq_device.copy_to_host()
            dolu = dolu_device.copy_to_host()
            docp = docp_device.copy_to_host()
        else:
            e_field_theta, e_field_phi = compute_electric_field_angle_components(
                self.simulation.numerics.lmax,
                self.simulation.parameters.particles.position,
                self.simulation.idx_lookup,
                self.simulation.scattered_field_coefficients,
                self.simulation.parameters.k_medium,
                self.simulation.numerics.azimuthal_angles,
                self.simulation.numerics.e_r,
                pilm,
                taulm,
            )

            intensity, dop, dolp, dolq, dolu, docp = compute_polarization_components(
                self.simulation.parameters.k_medium.size,
                self.simulation.numerics.azimuthal_angles.size,
                e_field_theta,
                e_field_phi,
            )

        self.scattering_angles = self.simulation.numerics.polar_angles

        self.phase_function_3d = (
            intensity
            * 4
            * np.pi
            / np.power(np.abs(self.simulation.parameters.k_medium), 2)
            / self.c_sca[np.newaxis, :]
        )
        self.phase_function_legendre_coefficients = np.polynomial.legendre.legfit(
            np.cos(self.scattering_angles),
            self.phase_function_3d,
            legendre_coefficients_number,
        )

        self.degree_of_polarization_3d = dop
        self.degree_of_linear_polarization_3d = dolp
        self.degree_of_linear_polarization_q_3d = dolq
        self.degree_of_linear_polarization_u_3d = dolu
        self.degree_of_circular_polarization_3d = docp

        if (self.simulation.numerics.sampling_points_number is not None) and (
            self.simulation.numerics.sampling_points_number.size == 2
        ):
            self.phase_function = np.mean(
                np.reshape(
                    self.phase_function_3d,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )

            self.degree_of_polarization = np.mean(
                np.reshape(
                    dop,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization = np.mean(
                np.reshape(
                    dolp,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization_q = np.mean(
                np.reshape(
                    dolq,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization_u = np.mean(
                np.reshape(
                    dolu,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_circular_polarization = np.mean(
                np.reshape(
                    docp,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )

            self.scattering_angles = np.reshape(
                self.scattering_angles, self.simulation.numerics.sampling_points_number
            )
            self.scattering_angles = self.scattering_angles[0, :]
        else:
            self.phase_function = self.phase_function_3d

            self.degree_of_polarization = dop
            self.degree_of_linear_polarization = dolp
            self.degree_of_linear_polarization_q = dolq
            self.degree_of_linear_polarization_u = dolu
            self.degree_of_circular_polarization = docp

        self.c_and_b_bounds = c_and_b
        if isinstance(c_and_b, bool):
            if c_and_b:
                self.c_and_b_bounds = ([-1, 0], [1, 1])
            else:
                return

        self.__compute_c_and_b()

    @staticmethod
    def compute_double_henyey_greenstein(theta: np.ndarray, cb: np.ndarray):
        cb = np.squeeze(cb)
        if cb.size < 2:
            cb = np.array([0, 0.5, 0.5])
        elif cb.size == 2:
            cb = np.append(cb, cb[1])
        elif cb.size > 3:
            cb = cb[:2]

        p1 = (1 - cb[1] ** 2) / np.power(
            1 - 2 * cb[1] * np.cos(theta) + cb[1] ** 2, 3 / 2
        )
        p2 = (1 - cb[2] ** 2) / np.power(
            1 + 2 * cb[2] * np.cos(theta) + cb[2] ** 2, 3 / 2
        )
        return (1 - cb[0]) / 2 * p1 + (1 + cb[0]) / 2 * p2

    def __compute_c_and_b(self):
        # double henyey greenstein
        if len(self.c_and_b_bounds[0]) not in [2, 3]:
            self.c_and_b_bounds = ([-1, 0], [1, 1])
            self.log.warning(
                "Number of parameters need to be 2 (b,c) or 3 (b1,b2,c). Reverting to two parameters (b,c) and setting the bounds to standard: b in [0, 1] and c in [-1, 1]"
            )

        from scipy.optimize import least_squares

        if len(self.c_and_b_bounds) == 2:
            bc0 = np.array([0, 0.5])
        else:
            bc0 = np.array([0, 0.5, 0.5])

        self.cb = np.empty((self.phase_function.shape[1], len(self.c_and_b_bounds)))
        for w in range(self.phase_function.shape[1]):

            def dhg_optimization(bc):
                return (
                    Optics.compute_double_henyey_greenstein(self.scattering_angles, bc)
                    - self.phase_function[:, w]
                )

            bc = least_squares(
                dhg_optimization, bc0, jac="2-point", bounds=self.c_and_b_bounds
            )
            self.cb[w, :] = bc.x



import yasfpy.log as log

import numpy as np
from scipy.sparse.linalg import LinearOperator, gmres, lgmres, bicgstab


class Solver:
    def __init__(
        self,
        solver_type: str = "gmres",
        tolerance: float = 1e-4,
        max_iter: int = 1e4,
        restart: int = 1e2,
    ):
        self.type = solver_type.lower()
        self.tolerance = tolerance
        self.max_iter = int(max_iter)
        self.restart = int(restart)

        self.log = log.scattering_logger(__name__)

    def run(self, a: LinearOperator, b: np.ndarray, x0: np.ndarray = None):
        if x0 is None:
            x0 = np.copy(b)

        if np.any(np.isnan(b)):
            print(b)

        if self.type == "bicgstab":
            counter = GMResCounter(callback_type="x")
            value, err_code = bicgstab(
                a,
                b,
                x0,
                tol=self.tolerance,
                atol=0,
                maxiter=self.max_iter,
                callback=counter,
            )
        elif self.type == "gmres":
            counter = GMResCounter(callback_type="pr_norm")
            value, err_code = gmres(
                a,
                b,
                x0,
                restart=self.restart,
                tol=self.tolerance,
                atol=self.tolerance**2,
                maxiter=self.max_iter,
                callback=counter,
                callback_type="pr_norm",
            )
        elif self.type == "lgmres":
            counter = GMResCounter(callback_type="x")
            value, err_code = lgmres(
                a,
                b,
                x0,
                tol=self.tolerance,
                atol=self.tolerance**2,
                maxiter=self.max_iter,
                callback=counter,
            )
        else:
            self.log.error("Please specify a valid solver type")
            exit(1)

        return value, err_code


class GMResCounter(object):
    def __init__(self, disp=False, callback_type="pr_norm"):
        self.log = log.scattering_logger(__name__)
        self._disp = disp
        self.niter = 0
        if callback_type == "pr_norm":
            self.header = "% 10s \t % 15s" % ("Iteration", "Residual")
        elif callback_type == "x":
            self.header = "% 10s \t %s" % ("Iteration", "Current Iterate")

    def __call__(self, rk=None):
        self.niter += 1
        if isinstance(rk, float):
            msg = "% 10i \t % 15.5f" % (self.niter, rk)
        elif isinstance(rk, np.ndarray):
            msg = "% 10i \t " % self.niter + np.array2string(rk)

        self.log.numerics(self.header)
        self.log.numerics(msg)
        if self._disp:
            print(self.header)
            print(msg)



import re
import numpy as np

from scipy.special import spherical_jn
from scipy.special import hankel1, lpmv

from yasfpy.functions.legendre_normalized_trigon import legendre_normalized_trigon


def jmult_max(num_part, lmax):
    return 2 * num_part * lmax * (lmax + 2)


def multi2single_index(j_s, tau, l, m, lmax):
    return (
        j_s * 2 * lmax * (lmax + 2)
        + (tau - 1) * lmax * (lmax + 2)
        + (l - 1) * (l + 1)
        + m
        + l
    )


def single_index2multi(idx, lmax):
    j_s = idx // (2 * lmax * (lmax + 2))
    idx_new = idx % (2 * lmax * (lmax + 2))
    tau = idx_new // (lmax * (lmax + 2)) + 1
    idx_new = idx_new % (lmax * (lmax + 2))
    l = np.floor(np.sqrt(idx_new + 1))
    m = idx_new - (l * l + l - 1)
    return j_s, tau, l, m


def transformation_coefficients(pilm, taulm, tau, l, m, pol, dagger: bool = False):
    ifac = 1j
    if dagger:
        ifac *= -1

    # Polarized light
    if np.any(np.equal(pol, [1, 2])):
        if tau == pol:
            spher_fun = taulm[l, np.abs(m)]
        else:
            spher_fun = m * pilm[l, np.abs(m)]

        return (
            -1
            / np.power(ifac, l + 1)
            / np.sqrt(2 * l * (l + 1))
            * (ifac * (pol == 1) + (pol == 2))
            * spher_fun
        )

    # Unpolarized light
    return (
        -1
        / np.power(ifac, l + 1)
        / np.sqrt(2 * l * (l + 1))
        * (ifac + 1)
        * (taulm[l, np.abs(m)] + m * pilm[l, np.abs(m)])
        / 2
    )


def mutual_lookup(
    lmax: int,
    positions_1: np.ndarray,
    positions_2: np.ndarray,
    refractive_index: np.ndarray,
    derivatives: bool = False,
    parallel: bool = False,
):
    differences = positions_1[:, np.newaxis, :] - positions_2[np.newaxis, :, :]
    distances = np.sqrt(np.sum(differences**2, axis=2))
    distances = distances[:, :, np.newaxis]
    e_r = np.divide(
        differences, distances, out=np.zeros_like(differences), where=distances != 0
    )
    cosine_theta = e_r[:, :, 2]
    # cosine_theta = np.divide(
    #   differences[:, :, 2],
    #   distances,
    #   out = np.zeros_like(distances),
    #   where = distances != 0
    # )
    # correct possible rounding errors
    cosine_theta[cosine_theta < -1] = -1
    cosine_theta[cosine_theta > 1] = 1
    sine_theta = np.sqrt(1 - cosine_theta**2)
    phi = np.arctan2(differences[:, :, 1], differences[:, :, 0])
    e_theta = np.stack(
        [cosine_theta * np.cos(phi), cosine_theta * np.sin(phi), -sine_theta], axis=-1
    )
    e_phi = np.stack([-np.sin(phi), np.cos(phi), np.zeros_like(phi)], axis=-1)

    size_parameter = distances * refractive_index[np.newaxis, np.newaxis, :]

    if parallel:
        from src.functions.cpu_numba import compute_lookup_tables

        spherical_bessel, spherical_hankel, e_j_dm_phi, p_lm = compute_lookup_tables(
            lmax, size_parameter, phi, cosine_theta
        )
    else:
        p_range = np.arange(2 * lmax + 1)
        p_range = p_range[:, np.newaxis, np.newaxis, np.newaxis]
        size_parameter_extended = size_parameter[np.newaxis, :, :, :]
        spherical_hankel = np.sqrt(
            np.divide(
                np.pi / 2,
                size_parameter_extended,
                out=np.zeros_like(size_parameter_extended),
                where=size_parameter_extended != 0,
            )
        ) * hankel1(p_range + 1 / 2, size_parameter_extended)
        spherical_bessel = spherical_jn(p_range, size_parameter_extended)

        if derivatives:
            spherical_hankel_lower = np.sqrt(
                np.divide(
                    np.pi / 2,
                    size_parameter_extended,
                    out=np.zeros_like(size_parameter_extended),
                    where=size_parameter_extended != 0,
                )
            ) * hankel1(-1 / 2, size_parameter_extended)
            spherical_hankel_lower = np.vstack(
                (spherical_hankel_lower, spherical_hankel[:-1, :, :, :])
            )
            spherical_hankel_derivative = (
                size_parameter_extended * spherical_hankel_lower
                - p_range * spherical_hankel
            )

            # p_range = np.arange(2 * lmax + 2) - 1
            # p_range = p_range[:, np.newaxis, np.newaxis, np.newaxis]
            # spherical_hankel = np.sqrt(np.divide(np.pi / 2, size_parameter_extended, out = np.zeros_like(size_parameter_extended), where = size_parameter_extended != 0)) * hankel1(p_range + 1/2, size_parameter_extended)
            # spherical_hankel_derivative = size_parameter_extended * spherical_hankel[:-1, :, :, :] - p_range[1:, :, :, :] * spherical_hankel[1:, :, :, :]

            p_lm = legendre_normalized_trigon(lmax, cosine_theta, sine_theta)
        else:
            spherical_hankel_derivative = None

            p_lm = np.zeros(
                (lmax + 1) * (2 * lmax + 1) * np.prod(size_parameter.shape[:2])
            ).reshape(((lmax + 1) * (2 * lmax + 1),) + size_parameter.shape[:2])
            for p in range(2 * lmax + 1):
                for absdm in range(p + 1):
                    cml = np.sqrt(
                        (2 * p + 1)
                        / 2
                        * np.prod(1 / np.arange(p - absdm + 1, p + absdm + 1))
                    )
                    # if np.isnan(cml):
                    #     print(p)
                    #     print(absdm)
                    p_lm[p * (p + 1) // 2 + absdm, :, :] = (
                        cml * np.power(-1.0, absdm) * lpmv(absdm, p, cosine_theta)
                    )

        phi = phi[np.newaxis, :, :]
        p_range = np.arange(-2 * lmax, 2 * lmax + 1)
        p_range = p_range[:, np.newaxis, np.newaxis]
        e_j_dm_phi = np.exp(1j * p_range * phi)

    return (
        spherical_bessel,
        spherical_hankel,
        e_j_dm_phi,
        p_lm,
        e_r,
        e_theta,
        e_phi,
        cosine_theta,
        sine_theta,
        size_parameter,
        spherical_hankel_derivative,
    )


import io
import pandas as pd
import yaml
import urllib.request
from urllib.parse import unquote
import requests as req


def material_handler(links):
    if not isinstance(links, list):
        links = [links]

    data = dict(ref_idx=pd.DataFrame(columns=["wavelength", "n", "k"]), material=None)
    for link in links:
        link = link.strip()
        if link[:4] == "http":
            if "refractiveindex.info" in link:
                df, material = handle_refractiveindex_info(link)
                data["ref_idx"] = pd.concat([data["ref_idx"], df])
                data["material"] = material
            elif "http://eodg.atm.ox.ac.uk" in link:
                df, material = handle_eodg(link)
                data["ref_idx"] = pd.concat([data["ref_idx"], df])
                data["material"] = material
            else:
                print("No mathing handler found for url")
        else:
            if ".csv" in link:
                df, material = handle_csv(link)
                data["ref_idx"] = pd.concat([data["ref_idx"], df])
                data["material"] = material
            else:
                print("No mathing handler found for file type")
    # data['ref_idx'] = data['ref_idx'].sort_values(by=['wavelength'])
    return data


def handle_refractiveindex_info(url):
    url_split = url.replace("=", "/").split("/")
    material = unquote(url_split[-2])

    if np.any([("data_csv" in part) or ("data_txt" in part) for part in url_split]):
        print("Please use the [Full database record] option for refractiveindex.info!")
        print("Reverting url:")
        print(f" from: {url}")
        url_split[3] = "database"
        url = "/".join(url_split)
        print(f" to:   {url}")

    # req = urllib.request.Request(url)
    #with urllib.request.urlopen(req) as resp:
    resp = req.get(url)
    if resp.status_code >= 400:
        raise Exception(f"Failed to retrieve data from {url}")
    # data = resp.read()
    # data = data.decode("utf-8")
    data = resp.text
    data_yml = yaml.safe_load(data)
    header_yml = ["wavelength", "n", "k"]
    data = pd.DataFrame(columns=["wavelength", "n", "k"])
    for line in data_yml["DATA"]:
        df = None
        if "tabulated" in line["type"].lower():
            # elif line['type'].lower()[-2:] == ' n':
            #   header_yml=['wavelength', 'n']
            if line["type"].lower()[-2:] == " k":
                header_yml = ["wavelength", "k", "n"]
            df = pd.read_csv(
                io.StringIO(line["data"]),
                delim_whitespace=True,
                header=None,
                names=header_yml,
            )
        elif "formula" in line["type"].lower():
            if line["type"].lower() == "formula 1":
                wavelengths = [float(c) for c in line["wavelength_range"].split()]
                wavelengths = np.arange(wavelengths[0], wavelengths[1], 0.1)
                coefficients = np.array(
                    [float(c) for c in line["coefficients"].split()]
                )
                ref_idx = lambda x: np.sqrt(
                    1
                    + np.sum(
                        [
                            coefficients[i]
                            * x**2
                            / (x**2 - coefficients[i + 1] ** 2)
                            for i in range(1, len(coefficients), 2)
                        ],
                        axis=0,
                    )
                )
                df = pd.DataFrame(columns=["wavelength", "n", "k"])
                df["wavelength"] = wavelengths
                df["n"] = ref_idx(wavelengths)

        if df is not None:
            df = df.fillna(0)
            data = pd.concat([data, df])

    return data, material


def handle_eodg(url):
    url_split = url.split("/")
    # material = unquote(url_split[-1][:-3]).replace('_', ' ')
    material = unquote(url_split[6])

    # req = urllib.request.Request(url)
    # with urllib.request.urlopen(req) as resp:
    # data = resp.read()
    # data = data.decode("iso-8859-1")
    resp = req.get(url)
    if resp.status_code >= 400:
        raise Exception(f"Failed to retrieve data from {url}")
    data = resp.text
    data_format = [
        s.lower() for s in re.search(r"#FORMAT=(.*)\n", data).group(1).split()
    ]
    header_yml = ["wavelength", "n", "k"]
    if "n" not in data_format:
        header_yml = ["wavelength", "k", "n"]

    data = re.sub(r"^#.*\n", "", data, flags=re.MULTILINE)
    data = pd.read_csv(
        io.StringIO(data), delim_whitespace=True, header=None, names=header_yml
    )
    data = data.fillna(0)
    # eodg uses wavenumbers in cm-1 instead of wavelengths in um, hence um = 1e4 / cm-1
    if "wavn" in data_format:
        data["wavelength"] = 1e4 / data["wavelength"]
        data = data.iloc[::-1]

    return data, material


def handle_csv(path):
    name = re.split("\._-", path)
    material = unquote(name[0])
    data = pd.read_csv(
        path, delim_whitespace=False, header=0, names=["wavelength", "n", "k"]
    )
    return data, material



import yasfpy.log as log

import numpy as np
from scipy.spatial import ConvexHull
from scipy.spatial.distance import pdist

from yasfpy.functions.misc import material_handler


class Particles:
    def __init__(
        self,
        position: np.array,
        r: np.array,
        refractive_index: np.array,
        refractive_index_table: list = None,
        type: str = "sphere",
    ):
        self.position = position
        self.r = r
        self.refractive_index = refractive_index
        self.type = type

        self.log = log.scattering_logger(__name__)

        # TODO: Keep it for now, remove later...
        self.refractive_index_table = refractive_index_table

        if refractive_index_table is None:
            if self.refractive_index.shape[1] == 2:
                self.refractive_index = (
                    self.refractive_index[:, 0] + 1j * self.refractive_index[:, 1]
                )
            elif self.refractive_index.shape[1] > 2:
                self.log.error(
                    "Refractive index should be either complex or a two column matrix!"
                )
        else:
            self.refractive_index = refractive_index.astype(int)
            self.refractive_index_table = refractive_index_table

        self.number = r.shape[0]
        self.__setup_impl()

    @staticmethod
    def generate_refractive_index_table(urls: list):
        data = [None] * len(urls)
        for k, url in enumerate(urls):
            data[k] = material_handler(url)

        return data

    def compute_unique_refractive_indices(self):
        self.unique_refractive_indices, self.refractive_index_array_idx = np.unique(
            self.refractive_index, return_inverse=True, axis=0
        )
        self.num_unique_refractive_indices = self.unique_refractive_indices.shape[0]

    def compute_unique_radii(self):
        self.unqiue_radii, self.radius_array_idx = np.unique(
            self.r, return_inverse=True, axis=0
        )
        self.num_unique_radii = self.unqiue_radii.shape[0]

    def compute_unique_radii_index_pairs(self):
        self.unique_radius_index_pairs, self.single_unique_array_idx = np.unique(
            np.column_stack((self.r, self.refractive_index)),
            return_inverse=True,
            axis=0,
        )
        self.unique_single_radius_index_pairs = np.unique(
            np.column_stack((self.radius_array_idx, self.refractive_index_array_idx)),
            axis=0,
        )

    def compute_single_unique_idx(self):
        self.single_unique_idx = (
            np.sum(self.unique_single_radius_index_pairs, axis=1)
            * (np.sum(self.unique_single_radius_index_pairs, axis=1) + 1)
        ) // 2 + self.unique_single_radius_index_pairs[:, 1]

        # pairedArray = (
        #   self.radius_array_idx + self.refractive_index_array_idx *
        #   (self.radius_array_idx + self.refractive_index_array_idx + 1)
        # ) // 2 + self.refractive_index_array_idx

        # self.single_unique_idx, self.single_unique_array_idx = np.unique(
        #   pairedArray,
        #   return_inverse=True,
        #   axis=0)

        self.num_unique_pairs = self.unique_radius_index_pairs.shape[0]

    def compute_maximal_particle_distance(self):
        hull = ConvexHull(self.position)
        vert = self.position[hull.vertices, :]
        self.max_particle_distance = max(pdist(vert))

    def compute_volume_equivalent_area(self):
        r3 = np.power(self.r, 3)
        self.geometric_projection = np.pi * np.power(np.sum(r3), 2 / 3)

    def __setup_impl(self):
        self.compute_unique_refractive_indices()
        self.compute_unique_radii()
        self.compute_unique_radii_index_pairs()
        self.compute_single_unique_idx()
        self.compute_maximal_particle_distance()
        self.compute_volume_equivalent_area()



import sys, os

sys.path.append(os.getcwd())
import yasfpy.log as log

from scipy.special import spherical_jn, spherical_yn


def t_entry(tau, l, k_medium, k_sphere, radius, field_type="scattered"):
    """
    Computes an entry in the T Matrix for a given l, k, and tau

    **Note**: scipy.special has also derivative function. Why is it not the same?
    Example:
      Now:    djx  = x *  spherical_jn(l-1, x)  - l * jx
      Possible: djx  = spherical_jn(l, x, derivative=True)
    """

    m = k_sphere / k_medium
    x = k_medium * radius
    mx = k_sphere * radius

    jx = spherical_jn(l, x)
    jmx = spherical_jn(l, mx)
    hx = spherical_jn(l, x) + 1j * spherical_yn(l, x)

    djx = x * spherical_jn(l - 1, x) - l * jx
    djmx = mx * spherical_jn(l - 1, mx) - l * jmx
    dhx = x * (spherical_jn(l - 1, x) + 1j * spherical_yn(l - 1, x)) - l * hx

    if (field_type, tau) == ("scattered", 1):
        return -(jmx * djx - jx * djmx) / (jmx * dhx - hx * djmx)  # -b
    elif (field_type, tau) == ("scattered", 2):
        return -(m**2 * jmx * djx - jx * djmx) / (
            m**2 * jmx * dhx - hx * djmx
        )  # -a
    elif (field_type, tau) == ("internal", 1):
        return (jx * dhx - hx * djx) / (jmx * dhx - hx * djmx)  # c
    elif (field_type, tau) == ("internal", 2):
        return (m * jx * dhx - m * hx * djx) / (m**2 * jmx * dhx - hx * djmx)  # d
    elif (field_type, tau) == ("ratio", 1):
        return (jx * dhx - hx * djx) / -(jmx * djx - jx * djmx)  # c / -b
    elif (field_type, tau) == ("ratio", 2):
        return (m * jx * dhx - m * hx * djx) / -(
            m**2 * jmx * djx - jx * djmx
        )  # d / -a
    else:
        logger = log.scattering_logger("t_entry")
        logger.warning("Not a valid field type provided. Returning None!")
        return None



from numba import jit, prange, complex128, float64, int64

import numpy as np
from scipy.special import spherical_jn, hankel1, lpmv


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, cache=True)
def particle_interaction(
    lmax: int,
    particle_number: int,
    idx: np.ndarray,
    x: np.ndarray,
    translation_table: np.ndarray,
    plm: np.ndarray,
    sph_h: np.ndarray,
    e_j_dm_phi,
):
    jmax = particle_number * 2 * lmax * (lmax + 2)
    channels = sph_h.shape[-1]

    wx = np.zeros(x.size * channels, dtype=complex128).reshape(x.shape + (channels,))

    for w_idx in prange(jmax * jmax * channels):
        w = w_idx % channels
        j_idx = w_idx // channels
        j1 = j_idx // jmax
        j2 = j_idx % jmax
        s1, n1, tau1, l1, m1 = idx[j1, :]
        s2, n2, tau2, l2, m2 = idx[j2, :]

        if s1 == s2:
            continue

        delta_tau = np.absolute(tau1 - tau2)
        delta_l = np.absolute(l1 - l2)
        delta_m = np.absolute(m1 - m2)

        val = 0j
        for p in range(np.maximum(delta_m, delta_l + delta_tau), l1 + l2 + 1):
            val += (
                translation_table[n2, n1, p]
                * plm[p * (p + 1) // 2 + delta_m, s1, s2]
                * sph_h[p, s1, s2, w]
            )
        val *= e_j_dm_phi[m2 - m1 + 2 * lmax, s1, s2] * x[j2]

        wx[j1, w] += val

    return wx


@jit(nopython=True, parallel=True, fastmath=True)
def compute_idx_lookups(lmax: int, particle_number: int):
    nmax = 2 * lmax * (lmax + 2)
    idx = np.zeros(nmax * particle_number * 5, dtype=int64).reshape(
        (nmax * particle_number, 5)
    )

    for s in prange(particle_number):
        for tau in range(1, 3):
            for l in range(1, lmax + 1):
                for m in range(-l, l + 1):
                    n = (tau - 1) * lmax * (lmax + 2) + (l - 1) * (l + 1) + l + m
                    i = n + s * nmax
                    idx[i, 0] = s
                    idx[i, 1] = n
                    idx[i, 2] = tau
                    idx[i, 3] = l
                    idx[i, 4] = m

    return idx


@jit(nopython=True, parallel=True, nogil=True, fastmath=True)
def compute_scattering_cross_section(
    lmax: int,
    particle_number: int,
    idx: np.ndarray,
    sfc: np.ndarray,
    translation_table: np.ndarray,
    plm: np.ndarray,
    sph_h: np.ndarray,
    e_j_dm_phi: np.ndarray,
):
    jmax = particle_number * 2 * lmax * (lmax + 2)
    channels = sph_h.shape[-1]

    c_sca_complex = np.zeros(channels, dtype=complex128)

    for j_idx in prange(jmax * jmax):
        j1 = j_idx // jmax
        j2 = j_idx % jmax
        s1, n1, _, _, m1 = idx[j1, :]
        s2, n2, _, _, m2 = idx[j2, :]

        delta_m = np.absolute(m1 - m2)

        p_dependent = np.zeros(channels, dtype=complex128)
        for p in range(delta_m, 2 * lmax + 1):
            p_dependent += (
                translation_table[n2, n1, p]
                * plm[p * (p + 1) // 2 + delta_m, s1, s2]
                * sph_h[p, s1, s2, :]
            )
        p_dependent *= (
            np.conj(sfc[s1, n1, :])
            * e_j_dm_phi[m2 - m1 + 2 * lmax, s1, s2]
            * sfc[s2, n2, :]
        )

        c_sca_complex += p_dependent

    return c_sca_complex


@jit(nopython=True, parallel=True, nogil=True, fastmath=True)
def compute_radial_independent_scattered_field_legacy(
    lmax: int,
    particles_position: np.ndarray,
    idx: np.ndarray,
    sfc: np.ndarray,
    k_medium: np.ndarray,
    azimuthal_angles: np.ndarray,
    e_r: np.ndarray,
    e_phi: np.ndarray,
    e_theta: np.ndarray,
    pilm: np.ndarray,
    taulm: np.ndarray,
):
    e_1_sca = np.zeros(
        azimuthal_angles.size * 3 * k_medium.size, dtype=complex128
    ).reshape((azimuthal_angles.size, 3, k_medium.size))
    jmax = particles_position.shape[0] * 2 * lmax * (lmax + 2)

    for global_idx in prange(jmax * azimuthal_angles.size * k_medium.size):
        w_idx = global_idx % (jmax * k_medium.size)
        g_idx = global_idx // (jmax * k_medium.size)

        a = g_idx

        w = w_idx % k_medium.size
        j_idx = w_idx // k_medium.size
        s, n, tau, l, m = idx[j_idx, :]

        t = (
            np.power(1j, tau - l - 2)
            * sfc[s, n, w]
            / np.sqrt(2 * l * (l + 1))
            * np.exp(
                1j
                * (
                    m * azimuthal_angles[a]
                    - k_medium[w] * np.sum(particles_position[s, :] * e_r[a, :])
                )
            )
        )

        if tau == 1:
            e_1_sca[a, :, w] += t * (
                e_theta[a, :] * pilm[l, np.abs(m), a] * 1j * m
                - e_phi[a, :] * taulm[l, np.abs(m), a]
            )
        else:
            e_1_sca[a, :, w] += t * (
                e_phi[a, :] * pilm[l, np.abs(m), a] * 1j * m
                + e_theta[a, :] * taulm[l, np.abs(m), a]
            )

    return e_1_sca


@jit(nopython=True, parallel=True, nogil=True, fastmath=True)
def compute_electric_field_angle_components(
    lmax: int,
    particles_position: np.ndarray,
    idx: np.ndarray,
    sfc: np.ndarray,
    k_medium: np.ndarray,
    azimuthal_angles: np.ndarray,
    e_r: np.ndarray,
    pilm: np.ndarray,
    taulm: np.ndarray,
):
    e_field_theta = np.zeros(
        azimuthal_angles.size * k_medium.size, dtype=complex128
    ).reshape((azimuthal_angles.size, k_medium.size))
    e_field_phi = np.zeros_like(e_field_theta)

    jmax = particles_position.shape[0] * 2 * lmax * (lmax + 2)

    for global_idx in prange(jmax * azimuthal_angles.size * k_medium.size):
        w_idx = global_idx % (jmax * k_medium.size)
        g_idx = global_idx // (jmax * k_medium.size)

        a = g_idx

        w = w_idx % k_medium.size
        j_idx = w_idx // k_medium.size
        s, n, tau, l, m = idx[j_idx, :]

        t = (
            np.power(1j, tau - l - 2)
            * sfc[s, n, w]
            / np.sqrt(2 * l * (l + 1))
            * np.exp(
                1j
                * (
                    m * azimuthal_angles[a]
                    - k_medium[w] * np.sum(particles_position[s, :] * e_r[a, :])
                )
            )
        )

        if tau == 1:
            e_field_theta[a, w] += t * pilm[l, np.abs(m), a] * 1j * m
            e_field_phi[a, w] -= t * taulm[l, np.abs(m), a]
        else:
            e_field_theta[a, w] += t * taulm[l, np.abs(m), a]
            e_field_phi[a, w] += t * pilm[l, np.abs(m), a] * 1j * m

    return e_field_theta, e_field_phi


@jit(nopython=True, parallel=True, nogil=True, fastmath=True)
def compute_polarization_components(
    number_of_wavelengths: int,
    number_of_angles: int,
    e_field_theta: np.ndarray,
    e_field_phi: np.ndarray,
):
    # Stokes components
    # S = np.zeros(4 * number_of_angles * number_of_wavelengths, dtype=complex128).reshape((4, number_of_angles, number_of_wavelengths))
    I = np.zeros(number_of_angles * number_of_wavelengths, dtype=float64).reshape(
        (number_of_angles, number_of_wavelengths)
    )
    Q = np.zeros_like(I)
    U = np.zeros_like(I)
    V = np.zeros_like(I)

    for global_idx in prange(number_of_angles * number_of_wavelengths):
        w_idx = global_idx % number_of_wavelengths
        a_idx = global_idx // number_of_wavelengths

        e_field_theta_abs = (
            e_field_theta[a_idx, w_idx].real ** 2
            + e_field_theta[a_idx, w_idx].imag ** 2
        )
        e_field_phi_abs = (
            e_field_phi[a_idx, w_idx].real ** 2 + e_field_phi[a_idx, w_idx].imag ** 2
        )
        e_field_angle_interaction = (
            e_field_theta[a_idx, w_idx] * e_field_phi[a_idx, w_idx].conjugate()
        )

        I[a_idx, w_idx] = e_field_theta_abs + e_field_phi_abs
        Q[a_idx, w_idx] = e_field_theta_abs - e_field_phi_abs
        U[a_idx, w_idx] = -2 * e_field_angle_interaction.real
        V[a_idx, w_idx] = 2 * e_field_angle_interaction.imag

    degree_of_polarization = np.sqrt(Q**2 + U**2 + V**2) / I
    degree_of_linear_polarization = np.sqrt(Q**2 + U**2) / I
    degree_of_linear_polarization_q = -Q / I
    degree_of_linear_polarization_u = U / I
    degree_of_circular_polarization = V / I

    return (
        I,
        degree_of_polarization,
        degree_of_linear_polarization,
        degree_of_linear_polarization_q,
        degree_of_linear_polarization_u,
        degree_of_circular_polarization,
    )


@jit(nopython=True, parallel=True, nogil=True, fastmath=True)
def compute_radial_independent_scattered_field(
    number_of_wavelengths: int,
    number_of_angles: int,
    e_phi: np.ndarray,
    e_theta: np.ndarray,
    e_field_theta: np.ndarray,
    e_field_phi: np.ndarray,
):
    e_1_sca = np.zeros(
        number_of_angles * 3 * number_of_wavelengths, dtype=complex128
    ).reshape((number_of_angles, 3, number_of_wavelengths))

    for global_idx in prange(number_of_angles * number_of_wavelengths):
        w = global_idx % number_of_wavelengths
        a = global_idx // number_of_wavelengths

        e_1_sca[a, :, w] = (
            e_field_theta[a, w] * e_theta[a, :] + e_field_phi[a, w] * e_phi[a, :]
        )

    return e_1_sca


@jit(nopython=False, parallel=True, nogil=False, fastmath=False, forceobj=True)
def compute_lookup_tables(
    lmax: int, size_parameter: np.ndarray, phi: np.ndarray, cosine_theta: np.ndarray
):
    spherical_hankel = np.zeros(
        (2 * lmax + 1) * np.prod(size_parameter.shape), dtype=complex
    ).reshape((2 * lmax + 1,) + size_parameter.shape)
    spherical_bessel = np.zeros_like(spherical_hankel)
    e_j_dm_phi = np.zeros(
        (4 * lmax + 1) * np.prod(size_parameter.shape[:2]), dtype=complex
    ).reshape((4 * lmax + 1,) + size_parameter.shape[:2])
    p_lm = np.zeros(
        (lmax + 1) * (2 * lmax + 1) * np.prod(size_parameter.shape[:2])
    ).reshape(((lmax + 1) * (2 * lmax + 1),) + size_parameter.shape[:2])

    for p in prange(2 * lmax + 1):
        spherical_hankel[p, :, :, :] = np.sqrt(
            np.divide(
                np.pi / 2,
                size_parameter,
                out=np.zeros_like(size_parameter),
                where=size_parameter != 0,
            )
        ) * hankel1(p + 1 / 2, size_parameter)
        spherical_bessel[p, :, :, :] = spherical_jn(p, size_parameter)
        e_j_dm_phi[p, :, :] = np.exp(1j * (p - 2 * lmax) * phi)
        e_j_dm_phi[p + 2 * lmax, :, :] = np.exp(1j * p * phi)
        for absdm in range(p + 1):
            cml = np.sqrt(
                (2 * p + 1) / 2 / np.prod(np.arange(p - absdm + 1, p + absdm + 1))
            )
            p_lm[p * (p + 1) // 2 + absdm, :, :] = (
                cml * np.power(-1.0, absdm) * lpmv(absdm, p, cosine_theta)
            )

    return spherical_bessel, spherical_hankel, e_j_dm_phi, p_lm


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, cache=True)
def compute_field(
    lmax: int,
    idx: np.ndarray,
    size_parameter: np.ndarray,
    sph_h: np.ndarray,
    derivative: np.ndarray,
    e_j_dm_phi: np.ndarray,
    p_lm: np.ndarray,
    pi_lm: np.ndarray,
    tau_lm: np.ndarray,
    e_r: np.ndarray,
    e_theta: np.ndarray,
    e_phi: np.ndarray,
    scattered_field_coefficients: np.ndarray = None,
    initial_field_coefficients: np.ndarray = None,
    scatter_to_internal: np.ndarray = None,
):
    jmax = sph_h.shape[1] * 2 * lmax * (lmax + 2)
    channels = sph_h.shape[-1]

    field = np.zeros(channels * sph_h.shape[2] * 3, dtype=complex128).reshape(
        (channels, sph_h.shape[2], 3)
    )

    if (scattered_field_coefficients is None) and (initial_field_coefficients is None):
        print(
            "At least one, scattered field or initial field coefficients, need to be given."
        )
        print("Returning a zero array")
        return field

    for w_idx in prange(2 * lmax * (lmax + 2) * np.prod(np.array(sph_h.shape[1:]))):
        w = w_idx % channels
        j_idx = w_idx // channels
        sampling_idx = j_idx // jmax
        j_idx = j_idx % jmax
        s, n, tau, l, m = idx[j_idx, :]

        invariant = (
            1 / np.sqrt(2 * (l + 1) * l) * e_j_dm_phi[m + 2 * lmax, s, sampling_idx]
        )
        # Calculate M
        if tau == 1:
            c_term_1 = (
                1j
                * m
                * pi_lm[l, np.abs(m), s, sampling_idx]
                * e_theta[s, sampling_idx, :]
            )
            c_term_2 = tau_lm[l, np.abs(m), s, sampling_idx] * e_phi[s, sampling_idx, :]
            c_term = sph_h[l, s, sampling_idx, w] * (c_term_1 - c_term_2)

            field[w, sampling_idx, :] += (
                scattered_field_coefficients[s, n, w] * invariant * c_term
            )
        # Calculate N
        else:
            p_term = (
                l
                * (l + 1)
                / size_parameter[s, sampling_idx, w]
                * sph_h[l, s, sampling_idx, w]
                * p_lm[l, np.abs(m), s, sampling_idx]
                * e_r[s, sampling_idx, :]
            )
            #   p_term = l * (l + 1) / size_parameter[s, sampling_idx, w]
            #   p_term *= sph_h[l, s, sampling_idx, w]
            #   p_term *= p_lm[l, np.abs(m), s, sampling_idx]
            #   p_term *= e_r[s, sampling_idx, :]

            b_term_1 = (
                derivative[l, s, sampling_idx, w] / size_parameter[s, sampling_idx, w]
            )
            b_term_2 = (
                tau_lm[l, np.abs(m), s, sampling_idx] * e_theta[s, sampling_idx, :]
            )
            b_term_3 = (
                1j
                * m
                * pi_lm[l, np.abs(m), s, sampling_idx]
                * e_phi[s, sampling_idx, :]
            )
            b_term = b_term_1 * (b_term_2 + b_term_3)

            field[w, sampling_idx, :] += (
                scattered_field_coefficients[s, n, w] * invariant * (p_term + b_term)
            )

    return field



import yasfpy.log as log

from typing import Union, Callable
import numpy as np
import pywigxjpf as wig

from yasfpy.functions.misc import jmult_max
from yasfpy.functions.misc import multi2single_index
from yasfpy.functions.legendre_normalized_trigon import legendre_normalized_trigon


class Numerics:
    def __init__(
        self,
        lmax: int,
        sampling_points_number: Union[int, np.ndarray] = 100,
        polar_angles: np.ndarray = None,
        polar_weight_func: Callable = lambda x: x,
        azimuthal_angles: np.ndarray = None,
        gpu: bool = False,
        particle_distance_resolution=10.0,
        solver=None,
    ):
        self.log = log.scattering_logger(__name__)
        self.lmax = lmax

        self.sampling_points_number = np.squeeze(sampling_points_number)

        if (polar_angles is None) or (azimuthal_angles is None):
            if self.sampling_points_number.size == 0:
                self.sampling_points_number = np.array([100])
                self.log.warning(
                    "Number of sampling points cant be an empty array. Reverting to 100 points (Fibonacci sphere)."
                )
            elif self.sampling_points_number.size > 2:
                self.sampling_points_number = np.array([sampling_points_number[0]])
                self.log.warning(
                    "Number of sampling points with more than two dimensions is not supported. Reverting to the first element in the provided array (Fibonacci sphere)."
                )

            if self.sampling_points_number.size == 1:
                (
                    _,
                    polar_angles,
                    azimuthal_angles,
                ) = Numerics.compute_fibonacci_sphere_points(sampling_points_number[0])
            elif self.sampling_points_number.size == 2:
                # if polar_weight_func is None:
                #   polar_weight_func = lambda x: x
                self.polar_angles_linspace = np.pi * polar_weight_func(
                    np.linspace(0, 1, sampling_points_number[1])
                )
                self.azimuthal_angles_linspace = (
                    2 * np.pi * np.linspace(0, 1, sampling_points_number[0] + 1)[:-1]
                )

                polar_angles, azimuthal_angles = np.meshgrid(
                    self.polar_angles_linspace,
                    self.azimuthal_angles_linspace,
                    indexing="xy",
                )

                polar_angles = polar_angles.ravel()
                azimuthal_angles = azimuthal_angles.ravel()

        else:
            self.sampling_points_number = None

        self.polar_angles = polar_angles
        self.azimuthal_angles = azimuthal_angles
        self.gpu = gpu
        self.particle_distance_resolution = particle_distance_resolution
        self.solver = solver

        if self.gpu:
            from numba import cuda

            if not cuda.is_available():
                self.log.warning(
                    "No supported GPU in numba detected! Falling back to the CPU implementation."
                )
                self.gpu = False

        self.__setup()

    def __compute_nmax(self):
        self.nmax = 2 * self.lmax * (self.lmax + 2)

    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.lpmn.html
    def __plm_coefficients(self):
        import sympy as sym

        self.plm_coeff_table = np.zeros(
            (2 * self.lmax + 1, 2 * self.lmax + 1, self.lmax + 1)
        )

        ct = sym.Symbol("ct")
        st = sym.Symbol("st")
        plm = legendre_normalized_trigon(2 * self.lmax, ct, y=st)

        for l in range(2 * self.lmax + 1):
            for m in range(l + 1):
                cf = sym.poly(plm[l, m], ct, st).coeffs()
                self.plm_coeff_table[l, m, 0 : len(cf)] = cf

    def __setup(self):
        self.__compute_nmax()
        # self.compute_translation_table()
        # self.__plm_coefficients()

    def compute_plm_coefficients(self):
        self.__plm_coefficients()

    def compute_translation_table(self):
        self.log.scatter("Computing the translation table")
        jmax = jmult_max(1, self.lmax)
        self.translation_ab5 = np.zeros((jmax, jmax, 2 * self.lmax + 1), dtype=complex)

        # No idea why or how this value for max_two_j works,
        # but got it through trial and error.
        # If you get any Wigner errors, change this value (e.g. 3*lmax)
        max_two_j = 3 * self.lmax
        wig.wig_table_init(max_two_j, 3)
        wig.wig_temp_init(max_two_j)

        # Needs to be paralilized or the loop needs to be shortened!
        # Probably using one/two loop(s) and index using the lookup table.
        for tau1 in range(1, 3):
            for l1 in range(1, self.lmax + 1):
                for m1 in range(-l1, l1 + 1):
                    j1 = multi2single_index(0, tau1, l1, m1, self.lmax)
                    for tau2 in range(1, 3):
                        for l2 in range(1, self.lmax + 1):
                            for m2 in range(-l2, l2 + 1):
                                j2 = multi2single_index(0, tau2, l2, m2, self.lmax)
                                for p in range(0, 2 * self.lmax + 1):
                                    if tau1 == tau2:
                                        self.translation_ab5[j1, j2, p] = (
                                            np.power(
                                                1j,
                                                abs(m1 - m2)
                                                - abs(m1)
                                                - abs(m2)
                                                + l2
                                                - l1
                                                + p,
                                            )
                                            * np.power(-1.0, m1 - m2)
                                            * np.sqrt(
                                                (2 * l1 + 1)
                                                * (2 * l2 + 1)
                                                / (2 * l1 * (l1 + 1) * l2 * (l2 + 1))
                                            )
                                            * (
                                                l1 * (l1 + 1)
                                                + l2 * (l2 + 1)
                                                - p * (p + 1)
                                            )
                                            * np.sqrt(2 * p + 1)
                                            * wig.wig3jj_array(
                                                2
                                                * np.array(
                                                    [l1, l2, p, m1, -m2, -m1 + m2]
                                                )
                                            )
                                            * wig.wig3jj_array(
                                                2 * np.array([l1, l2, p, 0, 0, 0])
                                            )
                                        )
                                    elif p > 0:
                                        self.translation_ab5[j1, j2, p] = (
                                            np.power(
                                                1j,
                                                abs(m1 - m2)
                                                - abs(m1)
                                                - abs(m2)
                                                + l2
                                                - l1
                                                + p,
                                            )
                                            * np.power(-1.0, m1 - m2)
                                            * np.sqrt(
                                                (2 * l1 + 1)
                                                * (2 * l2 + 1)
                                                / (2 * l1 * (l1 + 1) * l2 * (l2 + 1))
                                            )
                                            * np.lib.scimath.sqrt(
                                                (l1 + l2 + 1 + p)
                                                * (l1 + l2 + 1 - p)
                                                * (p + l1 - l2)
                                                * (p - l1 + l2)
                                                * (2 * p + 1)
                                            )
                                            * wig.wig3jj_array(
                                                2
                                                * np.array(
                                                    [l1, l2, p, m1, -m2, -m1 + m2]
                                                )
                                            )
                                            * wig.wig3jj_array(
                                                2 * np.array([l1, l2, p - 1, 0, 0, 0])
                                            )
                                        )

        wig.wig_table_free()
        wig.wig_temp_free()

    @staticmethod
    def compute_fibonacci_sphere_points(n=100):
        golden_ratio = (1 + 5**0.5) / 2
        i = np.arange(0, n)
        phi = 2 * np.pi * (i / golden_ratio % 1)
        theta = np.arccos(1 - 2 * i / n)

        return (
            np.stack(
                (
                    np.sin(theta) * np.cos(phi),
                    np.sin(theta) * np.sin(phi),
                    np.cos(theta),
                ),
                axis=1,
            ),
            theta,
            phi,
        )

    def compute_spherical_unity_vectors(self):
        self.e_r = np.stack(
            (
                np.sin(self.polar_angles) * np.cos(self.azimuthal_angles),
                np.sin(self.polar_angles) * np.sin(self.azimuthal_angles),
                np.cos(self.polar_angles),
            ),
            axis=1,
        )

        self.e_theta = np.stack(
            (
                np.cos(self.polar_angles) * np.cos(self.azimuthal_angles),
                np.cos(self.polar_angles) * np.sin(self.azimuthal_angles),
                -np.sin(self.polar_angles),
            ),
            axis=1,
        )

        self.e_phi = np.stack(
            (
                -np.sin(self.azimuthal_angles),
                np.cos(self.azimuthal_angles),
                np.zeros_like(self.azimuthal_angles),
            ),
            axis=1,
        )
